"""
Windows 2000 - List of Locale IDs and Language Groups

Original data table:
http://www.microsoft.com/globaldev/reference/win2k/setup/lcid.mspx
"""

LANGUAGE_ID = {
    0x0436: u"Afrikaans",
    0x041c: u"Albanian",
    0x0401: u"Arabic Saudi Arabia",
    0x0801: u"Arabic Iraq",
    0x0c01: u"Arabic Egypt",
    0x1001: u"Arabic Libya",
    0x1401: u"Arabic Algeria",
    0x1801: u"Arabic Morocco",
    0x1c01: u"Arabic Tunisia",
    0x2001: u"Arabic Oman",
    0x2401: u"Arabic Yemen",
    0x2801: u"Arabic Syria",
    0x2c01: u"Arabic Jordan",
    0x3001: u"Arabic Lebanon",
    0x3401: u"Arabic Kuwait",
    0x3801: u"Arabic UAE",
    0x3c01: u"Arabic Bahrain",
    0x4001: u"Arabic Qatar",
    0x042b: u"Armenian",
    0x042c: u"Azeri Latin",
    0x082c: u"Azeri Cyrillic",
    0x042d: u"Basque",
    0x0423: u"Belarusian",
    0x0402: u"Bulgarian",
    0x0403: u"Catalan",
    0x0404: u"Chinese Taiwan",
    0x0804: u"Chinese PRC",
    0x0c04: u"Chinese Hong Kong",
    0x1004: u"Chinese Singapore",
    0x1404: u"Chinese Macau",
    0x041a: u"Croatian",
    0x0405: u"Czech",
    0x0406: u"Danish",
    0x0413: u"Dutch Standard",
    0x0813: u"Dutch Belgian",
    0x0409: u"English United States",
    0x0809: u"English United Kingdom",
    0x0c09: u"English Australian",
    0x1009: u"English Canadian",
    0x1409: u"English New Zealand",
    0x1809: u"English Irish",
    0x1c09: u"English South Africa",
    0x2009: u"English Jamaica",
    0x2409: u"English Caribbean",
    0x2809: u"English Belize",
    0x2c09: u"English Trinidad",
    0x3009: u"English Zimbabwe",
    0x3409: u"English Philippines",
    0x0425: u"Estonian",
    0x0438: u"Faeroese",
    0x0429: u"Farsi",
    0x040b: u"Finnish",
    0x040c: u"French Standard",
    0x080c: u"French Belgian",
    0x0c0c: u"French Canadian",
    0x100c: u"French Swiss",
    0x140c: u"French Luxembourg",
    0x180c: u"French Monaco",
    0x0437: u"Georgian",
    0x0407: u"German Standard",
    0x0807: u"German Swiss",
    0x0c07: u"German Austrian",
    0x1007: u"German Luxembourg",
    0x1407: u"German Liechtenstein",
    0x0408: u"Greek",
    0x040d: u"Hebrew",
    0x0439: u"Hindi",
    0x040e: u"Hungarian",
    0x040f: u"Icelandic",
    0x0421: u"Indonesian",
    0x0410: u"Italian Standard",
    0x0810: u"Italian Swiss",
    0x0411: u"Japanese",
    0x043f: u"Kazakh",
    0x0457: u"Konkani",
    0x0412: u"Korean",
    0x0426: u"Latvian",
    0x0427: u"Lithuanian",
    0x042f: u"Macedonian",
    0x043e: u"Malay Malaysia",
    0x083e: u"Malay Brunei Darussalam",
    0x044e: u"Marathi",
    0x0414: u"Norwegian Bokmal",
    0x0814: u"Norwegian Nynorsk",
    0x0415: u"Polish",
    0x0416: u"Portuguese Brazilian",
    0x0816: u"Portuguese Standard",
    0x0418: u"Romanian",
    0x0419: u"Russian",
    0x044f: u"Sanskrit",
    0x081a: u"Serbian Latin",
    0x0c1a: u"Serbian Cyrillic",
    0x041b: u"Slovak",
    0x0424: u"Slovenian",
    0x040a: u"Spanish Traditional Sort",
    0x080a: u"Spanish Mexican",
    0x0c0a: u"Spanish Modern Sort",
    0x100a: u"Spanish Guatemala",
    0x140a: u"Spanish Costa Rica",
    0x180a: u"Spanish Panama",
    0x1c0a: u"Spanish Dominican Republic",
    0x200a: u"Spanish Venezuela",
    0x240a: u"Spanish Colombia",
    0x280a: u"Spanish Peru",
    0x2c0a: u"Spanish Argentina",
    0x300a: u"Spanish Ecuador",
    0x340a: u"Spanish Chile",
    0x380a: u"Spanish Uruguay",
    0x3c0a: u"Spanish Paraguay",
    0x400a: u"Spanish Bolivia",
    0x440a: u"Spanish El Salvador",
    0x480a: u"Spanish Honduras",
    0x4c0a: u"Spanish Nicaragua",
    0x500a: u"Spanish Puerto Rico",
    0x0441: u"Swahili",
    0x041d: u"Swedish",
    0x081d: u"Swedish Finland",
    0x0449: u"Tamil",
    0x0444: u"Tatar",
    0x041e: u"Thai",
    0x041f: u"Turkish",
    0x0422: u"Ukrainian",
    0x0420: u"Urdu",
    0x0443: u"Uzbek Latin",
    0x0843: u"Uzbek Cyrillic",
    0x042a: u"Vietnamese",
}

